# __init__.py
__version__ = "0.0.6"

"""
The :mod:`scifin.statistics` module includes methods for statistics.
"""

from .distributions  import Distribution, Normal, Uniform, Weibull, Rayleigh, Exponential, Gumbel, \
                            Laplace, Levy, Cauchy, \
                            Poisson, Binomial



